/**************************************************************/
/*                                                            */
/*  vge.h                                                     */
/*  Voxel Geometry Engine                                     */
/*  (c)1998 Marek Uher, xuher@sun.felk.cvut.cz                */
/*  http://service.felk.cvut.cz/~xuher/                       */
/*  Czech Technical University (CTU), Prague, Czech Republic  */
/*                                                            */
/*  vge v0.86 (November 1998)                                 */
/*                                                            */
/**************************************************************/

#ifndef VGE_H
#define VGE_H 1

#include <GL/gl.h>
/*                                                            */
/*  library version                                           */
/*                                                            */
#define VGE_VERSION "vge v0.86 (November 1998)"


/*                                                            */
/*  to debug                                                  */
/*                                                            */
#define DEBUG
/*                                                            */
/*  comment next line                                         */
/*                                                            */
#undef DEBUG


/*                                                            */
/*  texture filename                                          */
/*                                                            */
#define FILE_NAME 256


/*                                                            */
/*  Palette size (static palette)                             */
/*                                                            */
#define PALETTE_SIZE 256


/*                                                            */
/*  Direction for searching solid quads in voxel slice        */
/*                                                            */
#define ANY 0
#define RIGHT 1
#define DOWN 0
#define X_AND_Y 2


/*                                                            */
/*  Define auxiliary constant                                 */
/*                                                            */
#define NO 0
#define YES 1


/*                                                            */
/*  Error handling constant                                   */
/*                                                            */
#define OK 0
#define ERROR -1


/*                                                            */
/*  Symbolic constants for cutting voxel object               */
/*                                                            */
#define DIRECTION_X 0
#define DIRECTION_Y 1
#define DIRECTION_Z 2


/*                                                            */
/*  Constant for searching on voxel map                       */
/*                                                            */
#define CLEAN 0
#define DIRTY 1


/*                                                            */
/*  Transparent voxel                                         */
/*                                                            */
#define EMPTY_VOXEL 0


/*                                                            */
/*  Compressed or non-compressed data stream                  */
/*  (this version supported non-compressed data stream only)  */
/*                                                            */
#define NON_COMPRESED_DATA_STREAM 0
#define COMPRESED_DATA_STREAM 1


/*                                                            */
/*  Levels of detail                                          */
/*                                                            */
#define LOD_0 0
#define LOD_1 1
#define LOD_2 2
#define LOD_3 3
#define LOD_4 4


/*                                                            */
/*  Define new types                                          */
/*                                                            */


/*                                                            */
/*  Voxel entity   (cube 1x1x1)                               */
/*                                                            */
typedef int Voxel;


/*                                                            */
/*  Some value                                                */
/*                                                            */
typedef int Value;


/*                                                            */
/*  Resolution of voxel object, slice, map, ...               */
/*                                                            */
typedef int Resolution;


/*                                                            */
/*  Voxel color                                               */
/*                                                            */
typedef int Color;


/*                                                            */
/*  Voxel texture file version and release                    */
/*                                                            */
typedef int Version;
typedef int Release;


/*                                                            */
/*  General counter                                           */
/*                                                            */
typedef int Counter;


/*                                                            */
/*  General index                                             */
/*                                                            */
typedef int Index;


/*                                                            */
/*  Compress ratio (voxel faces - quad faces)                 */
/*                                                            */
typedef float Ratio;


/*                                                            */
/*  Color palette index                                       */
/*                                                            */
typedef int PaletteIndex;


/*                                                            */
/*  Level of detail                                           */
/*                                                            */
typedef int LOD;


/*                                                            */
/*  Flag (compressed/non-compressed data stream)              */
/*                                                            */
typedef int CDSflag;


/*                                                            */
/*  Direction for cutting voxel object                        */
/*                                                            */
typedef int Direction;


/*                                                            */
/*  Voxel slice, map, ...                                     */
/*                                                            */
typedef Voxel **VoxelSlice;


/*                                                            */
/*  General voxel buffer                                      */
/*                                                            */
typedef Voxel ***VoxelBuffer;


/*                                                            */
/*  Solid quad                                                */
/*                                                            */
typedef struct _Quad {
			Resolution x1;
			Resolution y1;
			Resolution x2;
			Resolution y2;
			Color color;
		} Quad;


/*                                                            */
/*  Color palette                                             */
/*                                                            */
typedef struct _ColorPalette {
		Color red;
		Color green;
		Color blue;
	} ColorPalette;


/*                                                            */
/*  Voxel object                                              */
/*                                                            */
typedef struct _VoxelObject {
		char objectName[FILE_NAME];
		Version version;
		Release release;
		Resolution resX;
		Resolution resY;
		Resolution resZ;
		LOD lod;
		CDSflag cdsFlag;
		ColorPalette colorPalette[PALETTE_SIZE];
		VoxelBuffer voxelBuffer;
	} VoxelObject;


/*                                                            */
/*  Error catching and handling                               */
/*                                                            */
#define POSIX_CLIB 1 
#if defined(POSIX_CLIB)
	/*                                                    */
	/*  Linux, Irix, ...                                  */
	/*                                                    */
	#define CatchException perror(sys_errlist[errno]); exit(errno)
#else
	/*                                                     */
	/*  SunOS, Solaris, ...                                */
	/*                                                     */
	#define CatchException fprintf(stderr, "%s\n", "critical error"); exit(-1)
#endif


/*                                                            */
/*  Release voxel buffer                                      */
/*                                                            */
#define ReleaseVoxelBuffer(voxBuffer) free((void *) voxBuffer)


/*                                                            */
/*  Release voxel slice                                       */
/*                                                            */
#define ReleaseVoxelSlice(voxSlice) free((void *) voxSlice)


/*                                                            */
/*  Write quad to file (debuging, normaly disabled)           */
/*                                                            */
void WriteQuad(Resolution x1, Resolution y1, Resolution z1, Resolution x2, Resolution y2, Resolution z2, Color color, const GLfloat *);


/*                                                            */
/*  Test if x is even or odd                                  */
/*                                                            */
int odd(int x);


/*                                                            */
/*  Function which initialize voxel slice to value            */
/*                                                            */
void InitVoxelSlice(VoxelSlice slice, Resolution x, Resolution y, Value value);


/*                                                            */
/*  Function which initialize voxel buffer to value           */
/*                                                            */
void InitVoxelBuffer(VoxelBuffer voxelBuffer, Resolution x, Resolution y, Resolution z, Value value);


/*                                                            */
/*  Function for creating voxel slice                         */
/*                                                            */  
VoxelSlice CreateVoxelSlice(Resolution x, Resolution y);


/*                                                            */
/*  Function for creating voxel buffer                        */
/*                                                            */
VoxelBuffer CreateVoxelBuffer(Resolution x, Resolution y, Resolution z);


/*                                                            */
/*  Function for creating voxel object in memory              */
/*                                                            */
VoxelObject *CreateVoxelObject(char *name, Version ver, Release rel, Voxel x, Voxel y, Voxel z, LOD lod, CDSflag cdsFlag, ColorPalette colorPalette[]);


/*                                                            */
/*  Release voxel object                                      */
/*                                                            */
void ReleaseVoxelObject(VoxelObject *voxObject);


/*                                                            */
/*  Parse texture file for tokens and data                    */
/*                                                            */
void ParseTextureFile(void);


/*                                                            */
/*  Find last clean index in voxel map                        */
/*                                                            */
void LastClean(Resolution *x, Resolution *y, Resolution maxI, Resolution maxJ, VoxelSlice map);


/*                                                            */
/*  Test if exist some clean position on voxel map            */
/*                                                            */
int IsClean(Resolution maxI, Resolution maxJ, VoxelSlice testedSlice);


/*                                                            */
/*  Mark index as dirty on index x,y in voxel map             */
/*                                                            */
void MarkDirty(Resolution x, Resolution y, VoxelSlice map);


/*                                                            */
/*  Find solid quads on voxel slice                           */
/*                                                            */ 
void FindQuad(Resolution x, Resolution y, Resolution maxX, Resolution maxY, VoxelSlice slice, VoxelSlice dirtyMap, Quad *quad);


/*                                                            */
/*  Convert voxel to solid faces                              */
/*                                                            */
void Voxel2Faces(Direction direction);


/*                                                            */
/*  Compute level of detail (LOD)                             */
/*                                                            */
void ComputeLOD(LOD nLod, char *objName);


/*                                                            */
/*  Convert voxel object to solid faces (B-rep)               */
/*                                                            */
GLuint ConvertVoxelObject2Faces(char *sourceFile, LOD newLod, int);
void getTextureSize(char *fn, Resolution *sx, Resolution *sy, Resolution *sz);


#endif


/* end of file vge.h */
