/***************************************************************************/
/*                                                                         */
/*                m3dt - 3D texture file mapper v0.23                      */
/*                  (c)1998 Marek Uher, uher@uhkt.cz                       */
/*                                                                         */
/*         txengine.c - module to b3dt (contain main texture engine)       */
/*                                                                         */
/***************************************************************************/

/* include header files */

/* standard system header files */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* user defined header files */
#include "parser.h"
#include "mpengine.h"

/* include header files specific for unix only */
#if defined(UNIX)
#include <sys/stat.h>
#include <unistd.h>
#endif               

/* global variables */

/* object info data structure */
struct {
	int x;
	int y;
	int sizeX;
	int sizeY;
	int placement;
	int boundary;
} objectInfo;	

/* input landscape data file */
char landscapeFileName[FILENAME_SIZE];

/* input texture container data description file */
char containerFileName[FILENAME_SIZE];

/* standard texture file header */
char fileHeader[SIZE_OF_FILE_HEADER];

/* debuging file */
FILE *debugFile;

/* input landscape file name handler */
FILE *landscapeFile;

/* output normalized landscape file */
FILE *normalizedLandscapeFile;

/* input texture container description file name handler */
FILE *containerFile;

/* object x-coordinate array */
int objectX[MAX_OBJECTS];

/* object y-coordinate array */
int objectY[MAX_OBJECTS];

/* size of object in voxels */ 
int sizeOfObjectX[MAX_OBJECTS];
int sizeOfObjectY[MAX_OBJECTS];
int sizeOfObjectZ[MAX_OBJECTS];

/* object type array */
int objectType[MAX_OBJECTS];

/* object placement flag */
int objectPlacement[MAX_OBJECTS];

/* object texture filename */
char *textureFileName[MAX_OBJECTS];

/* object texture file handler */
FILE *textureFile;

/* object placement messages */
char *strHorizontal = "horizontal";
char *strVertical = "vertical"; 

/* object type messages */
const char *strObjHighHouse = "high house";
const char *strObjBuildingHouse = "building house";
const char *strObjBoxHouse = "box house";
const char *strObjCountryHouse = "country house";
const char *strObjWaterWorks = "water works";
const char *strObjTree = "tree";
const char *strObjBush = "bush";
const char *strObjUnknow = "unknow object";

/* maximum of objects in scene */
int maxObjects = MAX_OBJECTS;

/* landscape data buffer */
int bufferCache[MAX][MAX]; 

/* normalized landscape data */
int normalizedCache[MAX][MAX];

/* set input landscape file name */
void SetLandscapeFileName(char *lndFileName){

	strcpy(landscapeFileName, lndFileName);
#if defined(DEBUG)
	printf("\n%s", landscapeFileName);
#endif
}

/* set input texture container description file name */
void SetContainerFileName(char *cntFileName){

        strcpy(containerFileName, cntFileName);
#if defined(DEBUG)
	printf("\n%s", containerFileName);
#endif
}

/* get parser error info */
void GetParserError(int objIndex, int message){

	printf("%c", '\a');
	switch(message){

		case OBJ_X_OVERFLOW:
			fprintf(stderr, "\n%s%d%s\n", "error: object #", objIndex," x-coordinates overflow");
			break;

		case OBJ_X_UNDERFLOW:
			fprintf(stderr, "\n%s%d%s\n", "error: object #", objIndex," x-coordinates underflow");
			break;

		case OBJ_Y_OVERFLOW:
			fprintf(stderr, "\n%s%d%s\n", "error: object #", objIndex," y-coordinates overflow");
			break;

		case OBJ_Y_UNDERFLOW:
			fprintf(stderr, "\n%s%d%s\n", "error: object #", objIndex," y-coordinates underflow");
			break;

		default:
			fprintf(stderr, "\n%s\n", "unknow error while parsing texture description file"); 
			break;
	}
}

/* test object for coordinates overlay */
void TestObjectPlacement(void){

	int index;
	int x;
	int y;
	int beginX;
	int beginY;
	int endX;
	int endY;
	int averageValue;
	int acumulator;
	int counter;

        objectInfo.boundary = BOUNDARY_SIZE;
        
        for(x = 0; x < MAX; x++){
        	for(y = 0; y < MAX; y++){
                	normalizedCache[x][y] = bufferCache[x][y];
                }
        }

        printf("\nopening normalized landscape data file: %s", NORMALIZED_LANDSCAPE_FILE);
        
        /* if some problem, say what's wrong and exit */
        if ((normalizedLandscapeFile = fopen(NORMALIZED_LANDSCAPE_FILE, "wa")) == NULL){
        	fprintf(stderr, "\n%s", "can't open output normalized landscape file");
                exit(ERROR);
        }
                                                                
        /* say, that there isn't problems with opening */
        printf("\n%s%s%s", "file ", NORMALIZED_LANDSCAPE_FILE, " successfully opened");
                                                                                
        for(index = 0; index < maxObjects; index++){
                                                                                                 
		objectInfo.x = objectX[index];
		objectInfo.y = objectY[index];
		
		objectInfo.sizeX = sizeOfObjectX[index];
		objectInfo.sizeY = sizeOfObjectY[index];

		objectInfo.placement = objectPlacement[index];

		/* test object coordinates for overflow or underflow */
		if (objectInfo.placement == HORIZONTAL_PLACEMENT){

			if (objectInfo.x < 0){
				GetParserError(index, OBJ_X_UNDERFLOW);
				exit(ERR_X_UNDERFLOW);
			}

			if ((objectInfo.x + objectInfo.sizeX) > MAX){
				GetParserError(index, OBJ_X_OVERFLOW);
				exit(ERR_X_OVERFLOW);
			}

			if (objectInfo.y < 0){
				GetParserError(index, OBJ_Y_UNDERFLOW);
				exit(ERR_Y_UNDERFLOW);
			}

			if ((objectInfo.y + objectInfo.sizeY) > MAX){
				GetParserError(index, OBJ_Y_OVERFLOW);
				exit(ERR_Y_OVERFLOW);
			}

			beginX = (objectInfo.x - objectInfo.boundary);
			if (beginX < 0) beginX = 0;

			beginY = (objectInfo.y - objectInfo.boundary);
			if (beginY < 0) beginY = 0;

			endX = (objectInfo.x + objectInfo.sizeX + objectInfo.boundary);
			if (endX > MAX) endX = MAX;

			endY = (objectInfo.y + objectInfo.sizeY + objectInfo.boundary);
			if (endY > MAX) endY = MAX;

			counter = 0;
			acumulator = 0;
			averageValue = 0;

			for(x = beginX; x <  endX; x++){
				for(y = beginY; y < endY; y++){
					acumulator += bufferCache[x][y];
					counter++;
				} 
			}

			averageValue = (int)((float)(acumulator)/(float)(counter));
		}

		if (objectInfo.placement == VERTICAL_PLACEMENT){

			if ((objectInfo.x - objectInfo.boundary) < 0){
				GetParserError(index, OBJ_X_UNDERFLOW);
				exit(ERR_X_UNDERFLOW);
			}

			if ((objectInfo.x + objectInfo.sizeY + objectInfo.boundary) > MAX){
				GetParserError(index, OBJ_X_OVERFLOW);
				exit(ERR_X_OVERFLOW);
			}

			if ((objectInfo.y - objectInfo.boundary) < 0){
				GetParserError(index, OBJ_Y_UNDERFLOW);
				exit(ERR_Y_UNDERFLOW);
			}

			if ((objectInfo.y + objectInfo.sizeX + objectInfo.boundary) > MAX){
				GetParserError(index, OBJ_Y_OVERFLOW);
				exit(ERR_Y_OVERFLOW);
			}

			beginX = (objectInfo.x - objectInfo.boundary);
			if (beginX < 0) beginX = 0;

			beginY = (objectInfo.y - objectInfo.boundary);
			if (beginY < 0) beginY = 0;

			endX = (objectInfo.x + objectInfo.sizeY + objectInfo.boundary);
			if (endX > MAX) endX = MAX;

			endY = (objectInfo.y + objectInfo.sizeX + objectInfo.boundary);
			if (endY > MAX) endY = MAX;

			counter = 0;
			acumulator = 0;
			averageValue = 0;

			for(x = beginX; x <  endX; x++){
				for(y = beginY; y < endY; y++){
					acumulator += bufferCache[x][y];
					counter++;
				} 
			}

			averageValue = (int)((float)(acumulator)/(float)(counter));
		}

		for(x = beginX; x < endX ;x++){
			for(y = beginY; y < endY ;y++){
				normalizedCache[x][y] = averageValue;
			}
		} 	
	}

	for(x = 0; x < MAX; x++){
		for(y = 0; y < MAX; y++){
			fprintf(normalizedLandscapeFile, "%d ", normalizedCache[x][y]);
		}
	}

	/* close file */
	if (fclose(normalizedLandscapeFile) != OK)

		/* if we have some problem with closeing file */
		printf("\n%s%s", "can't close output landscape file ", NORMALIZED_LANDSCAPE_FILE);

	else{
		printf("\n%s%s%s", "file ", NORMALIZED_LANDSCAPE_FILE, " closed");
		printf("%c", '\a');
	}	
}

/* map object with texture to scene */
void TestTextureFiles(void){

	int voxel;
	int counter;
	FILE *textureFile;
		
	printf("%s%d%s","Testing ", maxObjects, " texture file(s):"); 

	for(counter = 0; counter < maxObjects; counter++){
		printf("\nopening texture file: %s", textureFileName[counter]);

                /* if some problem, say what's wrong and exit */
                if ((textureFile = fopen(textureFileName[counter], "r")) == NULL){
                	PutMessage(M3DT_TFILE_ERR);
                        exit(ERROR);
                }

               /* say, that there isn't problems with opening */
               printf("\n%s%s%s", "file ", textureFileName[counter], " successfully opened");

               /* close file */
               if (fclose(textureFile) != OK)

               		/* if we have some problem with closeing file */
                        printf("\n%s%s", "cannot close input texture file ", textureFileName[counter]);
               else
               		printf("\n%s%s%s", "file ", textureFileName[counter], " closed");
	}
/*                
	printf("\n");
	fseek(containerFile, HEADER_OFFSET, SEEK_SET);
	while(!feof(containerFile)){
		fscanf(containerFile, "%d", &voxel);
		printf("%d", voxel);
	}
	printf("\n");
	fprintf(stdout, "%c", '\n');
*/
	fflush(stdout);
}

/* parse container file for objects, textures, placement */
/* and build internal scene database                     */
void ParseContainerFile(void){

	int x;
	int y;
	int sizeX;
	int sizeY;
	int sizeZ;
	int placement;
	char msgPlacement[MSG_SIZE];
	char msgObjectType[MSG_OBJ_SIZE];
	int index = 0;
	int counter = 0;
	int pause;
	char texFileName[FILENAME_SIZE];

	for(counter = 0; counter < MAX_OBJECTS; counter++)
		textureFileName[counter] = (char *) malloc(FILENAME_SIZE+ 1);		

	fseek(containerFile, CNT_OFFSET, SEEK_SET);

	while(!feof(containerFile)){
		printf("\n%s%d", "reading data for object: ", index);
		fscanf(containerFile, "%d %d %d %d %d %d %s", &x, &y, &placement, &sizeX, &sizeY, &sizeZ, &texFileName);
		
		objectX[index] = x;
		objectY[index] = y;
		
		strcpy(textureFileName[index], texFileName);

		if (placement == HORIZONTAL_PLACEMENT)
			strcpy(msgPlacement, strHorizontal);

		else{
			strcpy(msgPlacement, strVertical);
			placement = VERTICAL_PLACEMENT;
		}

		objectPlacement[index] = placement;

                sizeOfObjectX[index] = sizeX;
                sizeOfObjectY[index] = sizeY;
                sizeOfObjectZ[index] = sizeZ;
       		
       		printf("\n%s%d%s%d%s%d%s%d%s%d%s%s%s%s]", "object: x[", x, "], y[", y,"]\n\tsize of x[", sizeX,"], size of y[", sizeY,"], size of z[", sizeZ, "]\n\tplacement [", msgPlacement, "]\n\ttexture [", texFileName);                                
		/*
		printf("\n%s%s%s%d%s%d%s%s%s%s]", "object: [", msgObjectType, "]\n\tx[", x, "] y[", y, "]\n\tplacement [", msgPlacement, "]\n\ttexture [", texFileName);
		*/
		printf("\n%s", "--- Press enter to continue ---");

		pause = getchar();
		index++;			
	}
	maxObjects = index;
#if defined(DEBUG)
	for(counter = 0; counter < index; counter++)
		printf("\n%s", textureFileName[counter]);
#endif
	TestTextureFiles();
	TestObjectPlacement();
}

/* main texture function, which call */
/* one function on top               */
void Map3DTexture(void){

	int indexX;
	int indexY;
	int value;
	
	/* open output file */
        printf("\n%s%s", "opening debug file: ", DEBUG_FILE);

        /* if some problem, say what's wrong and exit */
        if ((debugFile = fopen(DEBUG_FILE, "wa")) == NULL){
		printf("\n%s%s", "Can't open debuging file: ", DEBUG_FILE);
        }

        /* say, that there isn't problems with opening */
        printf("\n%s%s%s", "file ", DEBUG_FILE, " successfully opened");
	
	
	/* open input landscape file */
        printf("\n%s%s", "opening input landscape file: ", landscapeFileName);

        /* if some problem, say what's wrong and exit */
        if ((landscapeFile = fopen(landscapeFileName, "r")) == NULL){
        	PutMessage(M3DT_LFILE_ERR);
                exit(ERROR);
        }

        /* say, that there isn't problems with opening */
        printf("\n%s%s%s", "file ", landscapeFileName, " successfully opened");

        /* open input texture container description file */
        printf("\n%s%s", "opening input texture container description file: ", containerFileName);

        /* if some problem, say what's wrong and exit */
        if ((containerFile = fopen(containerFileName, "r")) == NULL){
        	PutMessage(M3DT_CFILE_ERR);
                exit(ERROR);
        }                                                                        

        /* say, that there isn't problems with opening */
        printf("\n%s%s%s\n", "file ", containerFileName, " successfully opened");
	
	for(indexX = 0; indexX < MAX; indexX++){
		for(indexY = 0; indexY < MAX; indexY++){
			fscanf(landscapeFile, "%d", &bufferCache[indexX][indexY]);
			/*bufferCache[indexX][indexY] = value;*/
		} 
	}
	
#if defined(DEBUG)	
	for(indexX = 0; indexX < MAX; indexX++){
		for(indexY = 0; indexY < MAX; indexY++){
	        	fprintf(debugFile, "%d ", bufferCache[indexX][indexY]);
		}
	}
#endif	                                                                                               
	
	ParseContainerFile();
	
	rewind(landscapeFile);
	rewind(containerFile);

        /* close files */
        if (fclose(landscapeFile) != OK)

        	/* if we have some problem with closeing file */
                printf("\n%s", "cannot close input landscape file");
        else

                printf("\n%s%s%s", "file ", landscapeFileName, " closed");                        

        if (fclose(containerFile) != OK)

        	/* if we have some problem with closeing file */
                printf("\n%s", "cannot close input texture container description file");
        else

                printf("\n%s%s%s", "file ", containerFileName, " closed");        

        if (fclose(debugFile) != OK)
        	/* if we have some problem with closeing file */
                printf("\n%s", "cannot close debuging file");
        else
                printf("\n%s%s%s", "file ", DEBUG_FILE, " closed");        

}
                                                                                                                                                                                                                                                                                                                                                                                                                  
/* End of file: object.c */
